# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Recipes for Rich formatters and printers.

    .. note::

        To use this module, you must have the ``rich`` package installed.
'''



from rich.console import Console as _Console
from rich.pretty import pretty_repr as _pretty_repr

from . import __


_validate_arguments = (
    __.validate_arguments(
        globalvars = globals( ),
        errorclass = __.exceptions.ArgumentClassInvalidity ) )


class ConsoleTextIoInvalidity( __.exceptions.Omnierror, TypeError ):
    ''' Text stream invalid for use with Rich console. '''

    def __init__( self, stream: __.typx.Any ):
        super( ).__init__( f"Invalid stream for Rich console: {stream!r}" )


class Modes( __.enum.Enum ):
    ''' Operation modes for Rich truck. '''

    Formatter = 'formatter'
    Printer = 'printer'


ProduceTruckModeArgument: __.typx.TypeAlias = __.typx.Annotated[
    Modes,
    __.typx.Doc(
        ''' Operation mode.

            ``Formatter`` uses Rich to highlight and pretty text prior to
            printing (output). Text from non-Rich formatters will be printed
            as-is. Safer, but slightly more boring option.
            ``Printer`` uses Rich to highlight and pretty text while printing
            (output). Text from non-Rich formatters will be potentially be
            highlighted and prettied. If the text already contains ANSI SGR
            sequences (e.g., terminal colorization), then it might be
            reprocessed by the printer, causing visual artifacts. Less safe,
            but more vibrant option.
        ''' ),
]
ProduceTruckStderrArgument: __.typx.TypeAlias = __.typx.Annotated[
    bool, __.typx.Doc( ''' Output to standard diagnostic stream? ''' )
]


@_validate_arguments
def install( # noqa: PLR0913
    alias: __.InstallAliasArgument = __.builtins_alias_default,
    flavors: __.ProduceTruckFlavorsArgument = __.absent,
    active_flavors: __.ProduceTruckActiveFlavorsArgument = __.absent,
    trace_levels: __.ProduceTruckTraceLevelsArgument = __.absent,
    mode: ProduceTruckModeArgument = Modes.Formatter,
    stderr: ProduceTruckStderrArgument = True,
) -> __.Truck:
    ''' Produces truck and installs it into builtins with alias.

        Replaces an existing truck, preserving global module configurations.

        Library developers should call :py:func:`__.register_module` instead.
    '''
    truck = produce_truck(
        flavors = flavors,
        active_flavors = active_flavors,
        trace_levels = trace_levels,
        mode = mode,
        stderr = stderr )
    return truck.install( alias = alias )


@_validate_arguments
def produce_console_formatter(
    console: _Console,
    control: __.FormatterControl,
    mname: str,
    flavor: int | str,
) -> __.Formatter:
    ''' Produces formatter which uses Rich highlighter and prettier. '''
    return __.funct.partial( _console_format, console )


@_validate_arguments
def produce_console_printer(
    console: _Console, mname: str, flavor: __.Flavor
) -> __.Printer:
    # TODO: Remove from recipe. Should always use simple printer.
    ''' Produces a printer that uses Rich console printing.

        .. note::

            May reprocess ANSI SGR codes or markup from formatters, potentially
            causing visual artifacts. Be careful to use this only with "safe"
            formatters.
    '''
    return console.print


@_validate_arguments
def produce_pretty_formatter(
    control: __.FormatterControl, mname: str, flavor: int | str
) -> __.Formatter:
    ''' Produces formatter which uses Rich prettier. '''
    return _pretty_repr


@_validate_arguments
def produce_truck(
    flavors: __.ProduceTruckFlavorsArgument = __.absent,
    active_flavors: __.ProduceTruckActiveFlavorsArgument = __.absent,
    trace_levels: __.ProduceTruckTraceLevelsArgument = __.absent,
    mode: ProduceTruckModeArgument = Modes.Formatter,
    stderr: ProduceTruckStderrArgument = True,
) -> __.Truck:
    ''' Produces icecream truck which integrates with Rich. '''
    match mode:
        case Modes.Formatter: factory = _produce_formatter_truck
        case Modes.Printer: factory = _produce_printer_truck
    return factory(
        flavors = flavors,
        active_flavors = active_flavors,
        trace_levels = trace_levels,
        stderr = stderr )


@_validate_arguments
def register_module(
    name: __.RegisterModuleNameArgument = __.absent,
    flavors: __.ProduceTruckFlavorsArgument = __.absent,
    include_context: __.RegisterModuleIncludeContextArgument = __.absent,
    prefix_emitter: __.RegisterModulePrefixEmitterArgument = __.absent,
) -> None:
    ''' Registers module with Rich prettier to format arguments.

        Intended for library developers to configure debugging flavors without
        overriding anything set by the application or other libraries.
    '''
    __.register_module(
        name = name,
        flavors = flavors,
        formatter_factory = produce_pretty_formatter,
        include_context = include_context,
        prefix_emitter = prefix_emitter )


def _console_format( console: _Console, value: __.typx.Any ) -> str:
    with console.capture( ) as capture:
        console.print( value )
    return capture.get( )


def _produce_formatter_truck(
    flavors: __.ProduceTruckFlavorsArgument = __.absent,
    active_flavors: __.ProduceTruckActiveFlavorsArgument = __.absent,
    trace_levels: __.ProduceTruckTraceLevelsArgument = __.absent,
    stderr: ProduceTruckStderrArgument = True,
) -> __.Truck:
    console = _Console( stderr = stderr )
    gc_nomargs = { }
    if not __.is_absent( flavors ): gc_nomargs[ 'flavors' ] = flavors
    generalcfg = __.VehicleConfiguration(
        formatter_factory = __.funct.partial(
            produce_console_formatter, console ),
        **gc_nomargs ) # pyright: ignore
    target = __.sys.stderr if stderr else __.sys.stdout
    if not isinstance( target, __.io.TextIOBase ):
        raise ConsoleTextIoInvalidity( target )
    nomargs: dict[ str, __.typx.Any ] = dict(
        active_flavors = active_flavors,
        generalcfg = generalcfg,
        printer_factory = __.funct.partial(
            __.produce_simple_printer, target ),
        trace_levels = trace_levels )
    return __.produce_truck( **nomargs )


def _produce_printer_truck(
    flavors: __.ProduceTruckFlavorsArgument = __.absent,
    active_flavors: __.ProduceTruckActiveFlavorsArgument = __.absent,
    trace_levels: __.ProduceTruckTraceLevelsArgument = __.absent,
    stderr: ProduceTruckStderrArgument = True,
) -> __.Truck:
    console = _Console( stderr = stderr )
    gc_nomargs = { }
    if not __.is_absent( flavors ): gc_nomargs[ 'flavors' ] = flavors
    generalcfg = __.VehicleConfiguration(
        formatter_factory = produce_pretty_formatter,
        **gc_nomargs ) # pyright: ignore
    target = __.sys.stderr if stderr else __.sys.stdout
    if not isinstance( target, __.io.TextIOBase ): # pragma: no cover
        raise ConsoleTextIoInvalidity( target )
    nomargs: dict[ str, __.typx.Any ] = dict(
        active_flavors = active_flavors,
        generalcfg = generalcfg,
        printer_factory = __.funct.partial( produce_console_printer, console ),
        trace_levels = trace_levels )
    return __.produce_truck( **nomargs )


# def _produce_prefix( console: _Console, mname: str, flavor: _Flavor ) -> str:
#     # TODO: Detect if terminal supports 256 colors or true color.
#     #       Make spectrum of hues for trace depths, if so.
#     return _icecream.DEFAULT_PREFIX



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Common constants, imports, and utilities. '''


from .imports import *
from .miscellany import *
from .nomina import *
from .validators import *



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Common imports used throughout the package. '''

# ruff: noqa: F401


import collections.abc as   cabc
import contextlib as        ctxl
import dataclasses as       dcls
import                      enum
import functools as         funct
import                      inspect
import                      io
import itertools as         itert
import                      locale
import                      os
import                      re
import                      sys
import threading as         threads
import                      time
import                      types
import                      warnings

import accretive as         accret
import dynadoc as           ddoc
import frigid as            immut
import typing_extensions as typx
# --- BEGIN: Injected by Copier ---
# --- END: Injected by Copier ---

from absence import Absential, absent, is_absent



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Recipe for advanced formatters.

    .. note::

        To use this module, you must have the ``rich`` package installed.
'''

# TODO? Allow selection of trace color gradients.



from rich.console import Console as _Console
from rich.style import Style as _Style

from . import __


_validate_arguments = (
    __.validate_arguments(
        globalvars = globals( ),
        errorclass = __.exceptions.ArgumentClassInvalidity ) )


InterpolantsStylesRegistry: __.typx.TypeAlias = (
    __.accret.Dictionary[ str, _Style ] )


class Auxiliaries( __.immut.DataclassObject ):
    ''' Auxiliary functions used by formatters and interpolation.

        Typically used by unit tests to inject mock dependencies,
        but can also be used to deeply customize output.
    '''

    exc_info_discoverer: __.typx.Annotated[
        __.typx.Callable[ [ ], __.ExceptionInfo ],
        __.typx.Doc( ''' Returns information on current exception. ''' ),
    ] = __.sys.exc_info
    pid_discoverer: __.typx.Annotated[
        __.typx.Callable[ [ ], int ],
        __.typx.Doc( ''' Returns ID of current process. ''' ),
    ] = __.os.getpid
    thread_discoverer: __.typx.Annotated[
        __.typx.Callable[ [ ], __.threads.Thread ],
        __.typx.Doc( ''' Returns current thread. ''' ),
    ] = __.threads.current_thread
    time_formatter: __.typx.Annotated[
        __.typx.Callable[ [ str ], str ],
        __.typx.Doc( ''' Returns current time in specified format. ''' ),
    ] = __.time.strftime



class FlavorSpecification( __.immut.DataclassObject ):
    ''' Specification for custom flavor. '''

    color: __.typx.Annotated[
        str, __.typx.Doc( ''' Name of prefix color. ''' ) ]
    emoji: __.typx.Annotated[ str, __.typx.Doc( ''' Prefix emoji. ''' ) ]
    label: __.typx.Annotated[ str, __.typx.Doc( ''' Prefix label. ''' ) ]
    stack: __.typx.Annotated[
        bool, __.typx.Doc( ''' Include stack trace? ''' )
    ] = False


class PrefixDecorations( __.enum.IntFlag ):
    ''' Decoration styles for prefix emission. '''

    Plain =     0
    Color =     __.enum.auto( )
    Emoji =     __.enum.auto( )


class PrefixLabelPresentations( __.enum.IntFlag ):
    ''' How prefix label should be presented. '''

    Nothing =   0
    Words =     __.enum.auto( )
    Emoji =     __.enum.auto( )


class PrefixFormatControl( __.immut.DataclassObject ):
    ''' Format control for prefix emission. '''

    colorize: __.typx.Annotated[
        bool, __.typx.Doc( ''' Attempt to colorize? ''' )
    ] = True
    label_as: __.typx.Annotated[
        PrefixLabelPresentations,
        __.typx.Doc(
            ''' How to present prefix label.

                ``Words``: As words like ``TRACE0`` or ``ERROR``.
                ``Emoji``: As emoji like ``🔎`` or ``❌``.

                For both emoji and words: ``Emoji | Words``.
            ''' )
    ] = PrefixLabelPresentations.Words
    styles: __.typx.Annotated[
        InterpolantsStylesRegistry,
        __.typx.Doc(
            ''' Mapping of interpolant names to ``rich`` style objects. ''' ),
    ] = __.dcls.field( default_factory = InterpolantsStylesRegistry )
    template: __.typx.Annotated[
        str,
        __.typx.Doc(
            ''' String format for prefix.

                The following interpolants are supported:
                ``flavor``: Decorated flavor.
                ``module_qname``: Qualified name of invoking module.
                ``timestamp``: Current timestamp, formatted as string.
                ``process_id``: ID of current process according to OS kernel.
                ``thread_id``: ID of current thread.
                ``thread_name``: Name of current thread.
            ''' ),
    ] = "{flavor}| " # "{timestamp} [{module_qname}] {flavor}| "
    ts_format: __.typx.Annotated[
        str,
        __.typx.Doc(
            ''' String format for prefix timestamp.

                Used by :py:func:`time.strftime` or equivalent.
            ''' ),
    ] = '%Y-%m-%d %H:%M:%S.%f'


ProduceModulecfgAuxiliariesArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ Auxiliaries ],
    __.typx.Doc( ''' Auxiliary functions for formatting. ''' ),
]
ProduceModulecfgColorizeArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ bool ],
    __.typx.Doc( ''' Attempt to colorize output prefixes? ''' ),
]
ProduceModulecfgConsoleFactoryArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ __.typx.Callable[ [ ], _Console ] ],
    __.typx.Doc(
        ''' Factory function that produces Rich console instances. ''' ),
]
ProduceModulecfgPrefixLabelAsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ PrefixLabelPresentations ],
    __.typx.Doc(
        ''' How to present prefix labels (words, emoji, or both). ''' ),
]
ProduceModulecfgPrefixStylesArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ __.cabc.Mapping[ str, _Style ] ],
    __.typx.Doc( ''' Mapping of interpolant names to Rich style objects. ''' ),
]
ProduceModulecfgPrefixTemplateArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ str ],
    __.typx.Doc( ''' String template for prefix formatting. ''' ),
]
ProduceModulecfgPrefixTsFormatArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ str ],
    __.typx.Doc( ''' Timestamp format string for prefix. ''' ),
]


_flavor_specifications: __.immut.Dictionary[
    str, FlavorSpecification
] = __.immut.Dictionary(
    note = FlavorSpecification(
        color = 'blue',
        emoji = '\N{Information Source}\ufe0f',
        label = 'NOTE' ),
    monition = FlavorSpecification(
        color = 'yellow',
        emoji = '\N{Warning Sign}\ufe0f',
        label = 'MONITION' ),
    error = FlavorSpecification(
        color = 'red', emoji = '❌', label = 'ERROR' ),
    errorx = FlavorSpecification(
        color = 'red', emoji = '❌', label = 'ERROR', stack = True ),
    abort = FlavorSpecification(
        color = 'bright_red', emoji = '💥', label = 'ABORT' ),
    abortx = FlavorSpecification(
        color = 'bright_red', emoji = '💥', label = 'ABORT', stack = True ),
    future = FlavorSpecification(
        color = 'magenta', emoji = '🔮', label = 'FUTURE' ),
    success = FlavorSpecification(
        color = 'green', emoji = '✅', label = 'SUCCESS' ),
)

_flavor_aliases: __.immut.Dictionary[
    str, str
] = __.immut.Dictionary( {
    'n': 'note', 'm': 'monition',
    'e': 'error', 'a': 'abort',
    'ex': 'errorx', 'ax': 'abortx',
    'f': 'future', 's': 'success',
} )

_trace_color_names: tuple[ str, ... ] = (
    'grey85', 'grey82', 'grey78', 'grey74', 'grey70',
    'grey66', 'grey62', 'grey58', 'grey54', 'grey50' )

_trace_prefix_styles: tuple[ _Style, ... ] = tuple(
    _Style( color = name ) for name in _trace_color_names )


def _produce_console( ) -> _Console: # pragma: no cover
    # TODO? safe_box = True
    # Ideally, we want TTY so that Rich can detect proper attributes.
    # Failing that, stream to null device. (Output capture should still work.)
    for stream in ( __.sys.stderr, __.sys.stdout ):
        if not stream.isatty( ): continue
        return _Console( stderr = stream is __.sys.stderr )
    blackhole = open( # noqa: SIM115
        __.os.devnull, 'w', encoding = __.locale.getpreferredencoding( ) )
    # TODO? height = 24, width = 80
    return _Console( file = blackhole, force_terminal = True )


@_validate_arguments
def produce_module_configuration( # noqa: PLR0913
    colorize: ProduceModulecfgColorizeArgument = __.absent,
    prefix_label_as: ProduceModulecfgPrefixLabelAsArgument = __.absent,
    prefix_styles: ProduceModulecfgPrefixStylesArgument = __.absent,
    prefix_template: ProduceModulecfgPrefixTemplateArgument = __.absent,
    prefix_ts_format: ProduceModulecfgPrefixTsFormatArgument = __.absent,
    console_factory: ProduceModulecfgConsoleFactoryArgument = __.absent,
    auxiliaries: ProduceModulecfgAuxiliariesArgument = __.absent,
) -> __.ModuleConfiguration:
    ''' Produces module configuration with sundae-specific flavor settings. '''
    if __.is_absent( console_factory ): console_factory = _produce_console
    if __.is_absent( auxiliaries ): auxiliaries = Auxiliaries( )
    console = console_factory( )
    prefix_fmtctl_initargs: dict[ str, __.typx.Any ] = { }
    if not __.is_absent( colorize ):
        prefix_fmtctl_initargs[ 'colorize' ] = colorize
    if not __.is_absent( prefix_label_as ):
        prefix_fmtctl_initargs[ 'label_as' ] = prefix_label_as
    if not __.is_absent( prefix_styles ):
        prefix_fmtctl_initargs[ 'styles' ] = prefix_styles
    if not __.is_absent( prefix_template ):
        prefix_fmtctl_initargs[ 'template' ] = prefix_template
    if not __.is_absent( prefix_ts_format ):
        prefix_fmtctl_initargs[ 'ts_format' ] = prefix_ts_format
    prefix_fmtctl = PrefixFormatControl( **prefix_fmtctl_initargs )
    flavors = _produce_flavors( console, auxiliaries, prefix_fmtctl )
    formatter_factory = _produce_formatter_factory( console, auxiliaries )
    return __.ModuleConfiguration(
        flavors = flavors, formatter_factory = formatter_factory )


@_validate_arguments
def register_module( # noqa: PLR0913
    name: __.RegisterModuleNameArgument = __.absent,
    colorize: ProduceModulecfgColorizeArgument = __.absent,
    prefix_label_as: ProduceModulecfgPrefixLabelAsArgument = __.absent,
    prefix_styles: ProduceModulecfgPrefixStylesArgument = __.absent,
    prefix_template: ProduceModulecfgPrefixTemplateArgument = __.absent,
    prefix_ts_format: ProduceModulecfgPrefixTsFormatArgument = __.absent,
    console_factory: ProduceModulecfgConsoleFactoryArgument = __.absent,
    auxiliaries: ProduceModulecfgAuxiliariesArgument = __.absent,
) -> __.ModuleConfiguration:
    ''' Registers module with sundae-specific flavor configurations. '''
    configuration = produce_module_configuration(
        colorize = colorize,
        prefix_label_as = prefix_label_as,
        prefix_styles = prefix_styles,
        prefix_template = prefix_template,
        prefix_ts_format = prefix_ts_format,
        console_factory = console_factory,
        auxiliaries = auxiliaries )
    return __.register_module(
        name = name,
        flavors = configuration.flavors,
        formatter_factory = configuration.formatter_factory )


def _produce_flavors(
    console: _Console, auxiliaries: Auxiliaries, control: PrefixFormatControl
) -> __.FlavorsRegistry:
    emitter = _produce_prefix_emitter( console, auxiliaries, control )
    flavors: __.FlavorsRegistryLiberal = { }
    for name in _flavor_specifications:
        flavors[ name ] = __.FlavorConfiguration( prefix_emitter = emitter )
    for alias, name in _flavor_aliases.items( ):
        flavors[ alias ] = flavors[ name ]
    for level in range( 10 ):
        flavors[ level ] = __.FlavorConfiguration( prefix_emitter = emitter )
    return __.immut.Dictionary( flavors )


def _produce_formatter_factory(
    console: _Console, auxiliaries: Auxiliaries
) -> __.FormatterFactory:

    def factory(
        control: __.FormatterControl, mname: str, flavor: __.Flavor
    ) -> __.Formatter:

        def formatter( value: __.typx.Any ) -> str:
            tb_text = ''
            if isinstance( flavor, str ):
                flavor_ = _flavor_aliases.get( flavor, flavor )
                spec = _flavor_specifications[ flavor_ ]
                if spec.stack and auxiliaries.exc_info_discoverer( )[ 0 ]:
                    with console.capture( ) as capture:
                        console.print_exception( )
                    tb_text = capture.get( )
            else: flavor_ = flavor
            with console.capture( ) as capture:
                console.print( value, end = '' )
            text = capture.get( )
            if tb_text: return f"\n{tb_text}\n{text}"
            return text

        return formatter

    return factory


def _produce_prefix_emitter(
    console: _Console, auxiliaries: Auxiliaries, control: PrefixFormatControl
) -> __.PrefixEmitter:

    def emitter( mname: str, flavor: __.Flavor ) -> str:
        if isinstance( flavor, int ):
            return _produce_trace_prefix(
                console, auxiliaries, control, mname, flavor )
        name = _flavor_aliases.get( flavor, flavor )
        return _produce_special_prefix(
            console, auxiliaries, control, mname, name )

    return emitter


def _produce_special_prefix(
    console: _Console,
    auxiliaries: Auxiliaries,
    control: PrefixFormatControl,
    mname: str,
    flavor: str,
) -> str:
    styles = dict( control.styles )
    spec = _flavor_specifications[ flavor ]
    label = ''
    if control.label_as & PrefixLabelPresentations.Emoji:
        if control.label_as & PrefixLabelPresentations.Words:
            label = f"{spec.emoji} {spec.label}"
        else: label = f"{spec.emoji}"
    elif control.label_as & PrefixLabelPresentations.Words:
        label = f"{spec.label}"
    if control.colorize: styles[ 'flavor' ] = _Style( color = spec.color )
    return _render_prefix(
        console, auxiliaries, control, mname, label, styles )


def _produce_trace_prefix(
    console: _Console,
    auxiliaries: Auxiliaries,
    control: PrefixFormatControl,
    mname: str,
    level: int,
) -> str:
    # TODO? Option to render indentation guides.
    styles = dict( control.styles )
    label = ''
    if control.label_as & PrefixLabelPresentations.Emoji:
        if control.label_as & PrefixLabelPresentations.Words:
            label = f"🔎 TRACE{level}"
        else: label = '🔎'
    elif control.label_as & PrefixLabelPresentations.Words:
        label = f"TRACE{level}"
    if control.colorize and level < len( _trace_color_names ):
        styles[ 'flavor' ] = _Style( color = _trace_color_names[ level ] )
    indent = '  ' * level
    return _render_prefix(
        console, auxiliaries, control, mname, label, styles ) + indent


def _render_prefix( # noqa: PLR0913
    console: _Console,
    auxiliaries: Auxiliaries,
    control: PrefixFormatControl,
    mname: str,
    flavor: str,
    styles: dict[ str, _Style ],
) -> str:
    # TODO? Performance optimization: Only compute and interpolate PID, thread,
    #       and timestamp, if capabilities set permits.
    thread = auxiliaries.thread_discoverer( )
    interpolants: dict[ str, str ] = {
        'flavor': flavor,
        'module_qname': mname,
        'timestamp': auxiliaries.time_formatter( control.ts_format ),
        'process_id': str( auxiliaries.pid_discoverer( ) ),
        'thread_id': str( thread.ident ),
        'thread_name': thread.name,
    }
    if control.colorize: _stylize_interpolants( console, interpolants, styles )
    return control.template.format( **interpolants )


def _stylize_interpolants(
    console: _Console,
    interpolants: dict[ str, str ],
    styles: dict[ str, _Style ],
) -> None:
    style_default = styles.get( 'flavor' )
    interpolants_: dict[ str, str ] = { }
    for iname, ivalue in interpolants.items( ):
        style = styles.get( iname, style_default )
        if not style: continue # pragma: no branch
        with console.capture( ) as capture:
            console.print(
                ivalue, end = '', highlight = False, style = style  )
        interpolants_[ iname ] = capture.get( )
    interpolants.update( interpolants_ )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Printers, printer factories, and auxiliary functions and types. '''



import colorama as _colorama

from . import __
from . import configuration as _cfg
from . import exceptions as _exceptions


_validate_arguments = (
    __.validate_arguments(
        globalvars = globals( ),
        errorclass = _exceptions.ArgumentClassInvalidity ) )


Printer: __.typx.TypeAlias = __.cabc.Callable[ [ str ], None ]
PrinterFactory: __.typx.TypeAlias = (
    __.cabc.Callable[ [ str, _cfg.Flavor ], Printer ] )
PrinterFactoryUnion: __.typx.TypeAlias = __.io.TextIOBase | PrinterFactory


@_validate_arguments
def produce_simple_printer(
    target: __.io.TextIOBase,
    mname: str,
    flavor: _cfg.Flavor,
    force_color: bool = False,
) -> Printer:
    ''' Produces printer which uses standard Python 'print'. '''
    match __.sys.platform:
        case 'win32':
            winansi = _colorama.AnsiToWin32( target ) # pyright: ignore
            target_ = ( # pragma: no cover
                winansi.stream if winansi.convert else target )
        case _: target_ = target
    return __.funct.partial(
        _simple_print,
        target = target_, # pyright: ignore
        force_color = force_color )


def _remove_ansi_c1_sequences( text: str ) -> str:
    # https://stackoverflow.com/a/14693789/14833542
    regex = __.re.compile( r'''\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])''' )
    return regex.sub( '', text )


def _simple_print(
    text: str, target: __.io.TextIOBase, force_color = False
) -> None:
    if not force_color and not target.isatty( ):
        print( _remove_ansi_c1_sequences( text ), file = target )
        return
    print( text, file = target )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Miscellaneous package internals. '''



from . import imports as __


def install_builtin_safely(
    alias: str, installee: __.typx.Any, errorclass: type[ Exception ]
) -> None:
    ''' Adds attribute to interpreter builtins, if absent. '''
    import builtins
    if hasattr( builtins, alias ): raise errorclass( builtins, alias )
    setattr( builtins, alias, installee )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Family of exceptions for package API. '''


from . import __


class Omniexception(
    __.immut.Object, BaseException,
    instances_visibles = (
        '__cause__', '__context__', __.immut.is_public_identifier ),
):
    ''' Base for all exceptions raised by package API. '''


class Omnierror( Omniexception, Exception ):
    ''' Base for error exceptions raised by package API. '''


class ArgumentClassInvalidity( Omnierror, TypeError ):
    ''' Argument class is invalid. '''

    def __init__( self, name: str, classes: type | tuple[ type, ... ] ):
        if isinstance( classes, type ): classes = ( classes, )
        cnames = ' | '.join( map(
            lambda cls: f"{cls.__module__}.{cls.__qualname__}", classes ) )
        super( ).__init__(
            f"Argument {name!r} must be an instance of {cnames}." )


class AttributeNondisplacement( Omnierror, AttributeError ):
    ''' Cannot displace existing attribute. '''

    def __init__( self, object_: __.typx.Any, name: str ):
        super( ).__init__(
            f"Cannot displace attribute {name!r} on: {object_}" )


class FlavorInavailability( Omnierror, ValueError ):
    ''' Requested flavor is not available. '''

    def __init__( self, flavor: int | str ):
        super( ).__init__( f"Flavor {flavor!r} is not available." )


class ModuleInferenceFailure( Omnierror, RuntimeError ):
    ''' Failure to infer invoking module from call stack. '''

    def __init__( self ):
        super( ).__init__( "Could not infer invoking module from call stack." )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Validators for internal use. '''



from . import imports as __


def validate_arguments(
    globalvars: dict[ str, __.typx.Any ],
    errorclass: type[ Exception ],
):
    ''' Decorator factory which produces argument validators. '''

    def decorate( function: __.cabc.Callable[ ..., __.typx.Any ] ):
        ''' Decorates function to be validated. '''

        @__.funct.wraps( function )
        def validate( *posargs: __.typx.Any, **nomargs: __.typx.Any ):
            ''' Validates arguments before invocation. '''
            signature = __.inspect.signature( function )
            inspectee = signature.bind( *posargs, **nomargs )
            inspectee.apply_defaults( )
            for name, value in inspectee.arguments.items( ):
                param = signature.parameters[ name ]
                annotation = param.annotation
                if __.is_absent( value ): continue
                if annotation is param.empty: continue
                classes = _reduce_annotation(
                    annotation, globalvars = globalvars )
                if not isinstance( value, classes ):
                    raise errorclass( name, classes )
            return function( *posargs, **nomargs )

        return validate

    return decorate


def _reduce_annotation(
    annotation: __.typx.Any, globalvars: dict[ str, __.typx.Any ]
) -> tuple[ type, ... ]:
    if isinstance( annotation, str ):
        return _reduce_annotation(
            eval( annotation, globalvars ), # noqa: S307
            globalvars = globalvars )
    origin = __.typx.get_origin( annotation )
    if isinstance( annotation, __.types.UnionType ) or origin is __.typx.Union:
        return tuple( __.itert.chain.from_iterable(
            map(
                lambda a: _reduce_annotation( a, globalvars = globalvars ),
                __.typx.get_args( annotation ) ) ) )
    if origin is None: return ( annotation, )
    if origin is __.typx.Annotated:
        return _reduce_annotation(
            annotation.__origin__, globalvars = globalvars )
    # TODO? Other special forms.
    return ( origin, )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Portions of configuration hierarchy. '''


import icecream as _icecream

from . import __


class FormatterControl( __.immut.DataclassObject ):
    ''' Contextual data for formatter and prefix factories. '''

    columns_count_effective: __.typx.Annotated[
        __.typx.Optional[ int ],
        __.typx.Doc(
            ''' Available line length after accounting for embellishments.

                May be ``None`` if not determinable.
            '''
        ),
    ] = None


Flavor: __.typx.TypeAlias = int | str
Formatter: __.typx.TypeAlias = __.typx.Callable[ [ __.typx.Any ], str ]
FormatterFactory: __.typx.TypeAlias = (
    __.typx.Callable[ [ FormatterControl, str, Flavor ], Formatter ] )
PrefixEmitter: __.typx.TypeAlias = __.typx.Callable[ [ str, Flavor ], str ]
PrefixEmitterUnion: __.typx.TypeAlias = str | PrefixEmitter


class FlavorConfiguration( __.immut.DataclassObject ):
    ''' Per-flavor configuration. '''

    formatter_factory: __.typx.Annotated[
        __.typx.Optional[ FormatterFactory ],
        __.typx.Doc(
            ''' Factory which produces formatter callable.

                Takes formatter control, module name, and flavor as arguments.
                Returns formatter to convert an argument to a string.

                Default ``None`` inherits from cumulative configuration.
            ''' ),
    ] = None
    include_context: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.typx.Doc(
            ''' Include stack frame with output?

                Default ``None`` inherits from cumulative configuration.
            ''' ),
    ] = None
    prefix_emitter: __.typx.Annotated[
        __.typx.Optional[ PrefixEmitterUnion ],
        __.typx.Doc(
            ''' String or factory which produces output prefix string.

                Factory takes formatter control, module name, and flavor as
                arguments. Returns prefix string.

                Default ``None`` inherits from cumulative configuration.
            ''' ),
    ] = None


def produce_default_flavors( ) -> __.immut.Dictionary[
    Flavor, FlavorConfiguration
]:
    ''' Produces flavors for trace depths 0 through 9. '''
    return __.immut.Dictionary( {
        i: FlavorConfiguration(
            prefix_emitter = f"TRACE{i}| " ) for i in range( 10 ) } )


FlavorsRegistry: __.typx.TypeAlias = (
    __.immut.Dictionary[ Flavor, FlavorConfiguration ] )
FlavorsRegistryLiberal: __.typx.TypeAlias = (
    __.cabc.Mapping[ Flavor, FlavorConfiguration ] )


class ModuleConfiguration( __.immut.DataclassObject ):
    ''' Per-module or per-package configuration. '''

    flavors: __.typx.Annotated[
        FlavorsRegistry,
        __.typx.Doc(
            ''' Registry of flavor identifiers to configurations. ''' ),
    ] = __.dcls.field( default_factory = FlavorsRegistry )
    formatter_factory: __.typx.Annotated[
        __.typx.Optional[ FormatterFactory ],
        __.typx.Doc(
            ''' Factory which produces formatter callable.

                Takes formatter control, module name, and flavor as arguments.
                Returns formatter to convert an argument to a string.

                Default ``None`` inherits from cumulative configuration.
            ''' ),
    ] = None
    include_context: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.typx.Doc(
            ''' Include stack frame with output?

                Default ``None`` inherits from cumulative configuration.
            ''' ),
    ] = None
    prefix_emitter: __.typx.Annotated[
        __.typx.Optional[ PrefixEmitterUnion ],
        __.typx.Doc(
            ''' String or factory which produces output prefix string.

                Factory takes formatter control, module name, and flavor as
                arguments. Returns prefix string.

                Default ``None`` inherits from cumulative configuration.
            ''' ),
    ] = None


class VehicleConfiguration( __.immut.DataclassObject ):
    ''' Per-vehicle configuration. '''

    flavors: __.typx.Annotated[
        FlavorsRegistry,
        __.typx.Doc(
            ''' Registry of flavor identifiers to configurations. ''' ),
    ] = __.dcls.field( default_factory = produce_default_flavors )
    formatter_factory: __.typx.Annotated[
        FormatterFactory,
        __.typx.Doc(
            ''' Factory which produces formatter callable.

                Takes formatter control, module name, and flavor as arguments.
                Returns formatter to convert an argument to a string.
            ''' ),
    ] = lambda ctrl, mname, flavor: _icecream.DEFAULT_ARG_TO_STRING_FUNCTION
    include_context: __.typx.Annotated[
        bool, __.typx.Doc( ''' Include stack frame with output? ''' )
    ] = False
    prefix_emitter: __.typx.Annotated[
        PrefixEmitterUnion,
        __.typx.Doc(
            ''' String or factory which produces output prefix string.

                Factory takes formatter control, module name, and flavor as
                arguments. Returns prefix string.
            ''' ),
    ] = _icecream.DEFAULT_PREFIX



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Vehicles which vend flavors of Icecream debugger. '''



import icecream as _icecream

from . import __
from . import configuration as _cfg
from . import exceptions as _exceptions
from . import printers as _printers


# if __.typx.TYPE_CHECKING: # pragma: no cover
#     import _typeshed


_installer_lock: __.threads.Lock = __.threads.Lock( )
_registrar_lock: __.threads.Lock = __.threads.Lock( )
_self_modulecfg: _cfg.ModuleConfiguration = _cfg.ModuleConfiguration(
    flavors = __.immut.Dictionary(
        note = _cfg.FlavorConfiguration( prefix_emitter = 'NOTE| ' ),
        error = _cfg.FlavorConfiguration( prefix_emitter = 'ERROR| ' ) ) )
_validate_arguments = (
    __.validate_arguments(
        globalvars = globals( ),
        errorclass = _exceptions.ArgumentClassInvalidity ) )


class ModulesConfigurationsRegistry(
    __.accret.Dictionary[ str, _cfg.ModuleConfiguration ]
):
    ''' Accretive dictionary specifically for module registrations. '''

    def __init__(
        self,
        *iterables: __.DictionaryPositionalArgument[
            str, _cfg.ModuleConfiguration ],
        **entries: __.DictionaryNominativeArgument[
            _cfg.ModuleConfiguration ],
    ):
        super( ).__init__( { __.package_name: _self_modulecfg } )
        self.update( *iterables, **entries )


class Omniflavor( __.enum.Enum ):
    ''' Singleton to match any flavor. '''

    Instance = __.enum.auto( )


ActiveFlavors: __.typx.TypeAlias = Omniflavor | frozenset[ _cfg.Flavor ]
ActiveFlavorsLiberal: __.typx.TypeAlias = __.typx.Union[
    Omniflavor,
    __.cabc.Sequence[ _cfg.Flavor ],
    __.cabc.Set[ _cfg.Flavor ],
]
ActiveFlavorsRegistry: __.typx.TypeAlias = (
    __.immut.Dictionary[ str | None, ActiveFlavors ] )
ActiveFlavorsRegistryLiberal: __.typx.TypeAlias = (
    __.cabc.Mapping[ str | None, ActiveFlavorsLiberal ] )
ModulesConfigurationsRegistryLiberal: __.typx.TypeAlias = (
    __.cabc.Mapping[ str, _cfg.ModuleConfiguration ] )
ReportersRegistry: __.typx.TypeAlias = (
    __.accret.Dictionary[
        tuple[ str, _cfg.Flavor ], _icecream.IceCreamDebugger ] )
TraceLevelsRegistry: __.typx.TypeAlias = (
    __.immut.Dictionary[ str | None, int ] )
TraceLevelsRegistryLiberal: __.typx.TypeAlias = (
    __.cabc.Mapping[ str | None, int ] )


builtins_alias_default: __.typx.Annotated[
    str,
    __.typx.Doc( ''' Default alias for global truck in builtins module. ''' ),
] = 'ictr'
modulecfgs: __.typx.Annotated[
    ModulesConfigurationsRegistry,
    __.typx.Doc( ''' Global registry of module configurations. ''' ),
] = ModulesConfigurationsRegistry( )
omniflavor: __.typx.Annotated[
    Omniflavor, __.typx.Doc( ''' Matches any flavor. ''' )
] = Omniflavor.Instance


class Truck( __.immut.DataclassObject ):
    ''' Vends flavors of Icecream debugger. '''

    active_flavors: __.typx.Annotated[
        ActiveFlavorsRegistry,
        __.typx.Doc(
            ''' Mapping of module names to active flavor sets.

                Key ``None`` applies globally. Module-specific entries
                override globals for that module.
            ''' ),
    ] = __.dcls.field( default_factory = ActiveFlavorsRegistry )
    generalcfg: __.typx.Annotated[
        _cfg.VehicleConfiguration,
        __.typx.Doc(
            ''' General configuration.

                Top of configuration inheritance hierarchy.
                Default is suitable for application use.
            ''' ),
    ] = __.dcls.field( default_factory = _cfg.VehicleConfiguration )
    modulecfgs: __.typx.Annotated[
        ModulesConfigurationsRegistry,
        __.typx.Doc(
            ''' Registry of per-module configurations.

                Modules inherit configuration from their parent packages.
                Top-level packages inherit from general instance
                configruration.
            ''' ),
    ] = __.dcls.field( default_factory = lambda: modulecfgs )
    printer_factory: __.typx.Annotated[
        _printers.PrinterFactoryUnion,
        __.typx.Doc(
            ''' Factory which produces callables to output text somewhere.

                May also be writable text stream.
                Factories take two arguments, module name and flavor, and
                return a callable which takes one argument, the string
                produced by a formatter.
            ''' ),
    ] = __.funct.partial( _printers.produce_simple_printer, __.sys.stderr )
    trace_levels: __.typx.Annotated[
        TraceLevelsRegistry,
        __.typx.Doc(
            ''' Mapping of module names to maximum trace depths.

                Key ``None`` applies globally. Module-specific entries
                override globals for that module.
            ''' ),
    ] = __.dcls.field(
        default_factory = lambda: __.immut.Dictionary( { None: -1 } ) )
    _debuggers: __.typx.Annotated[
        ReportersRegistry,
        __.typx.Doc(
            ''' Cache of debugger instances by module and flavor. ''' ),
    ] = __.dcls.field( default_factory = ReportersRegistry )
    _debuggers_lock: __.typx.Annotated[
        __.threads.Lock,
        __.typx.Doc( ''' Access lock for cache of debugger instances. ''' ),
    ] = __.dcls.field( default_factory = __.threads.Lock )

    @_validate_arguments
    def __call__(
        self,
        flavor: _cfg.Flavor, *,
        module_name: __.Absential[ str ] = __.absent,
    ) -> _icecream.IceCreamDebugger:
        ''' Vends flavor of Icecream debugger. '''
        mname = (
            _discover_invoker_module_name( ) if __.is_absent( module_name )
            else module_name )
        cache_index = ( mname, flavor )
        if cache_index in self._debuggers:
            with self._debuggers_lock:
                return self._debuggers[ cache_index ]
        configuration = _produce_ic_configuration( self, mname, flavor )
        control = _cfg.FormatterControl( )
        initargs = _calculate_ic_initargs(
            self, configuration, control, mname, flavor )
        debugger = _icecream.IceCreamDebugger( **initargs )
        if isinstance( flavor, int ):
            trace_level = (
                _calculate_effective_trace_level( self.trace_levels, mname) )
            debugger.enabled = flavor <= trace_level
        elif isinstance( flavor, str ): # pragma: no branch
            active_flavors = (
                _calculate_effective_flavors( self.active_flavors, mname ) )
            debugger.enabled = (
                isinstance( active_flavors, Omniflavor )
                or flavor in active_flavors )
        with self._debuggers_lock:
            self._debuggers[ cache_index ] = debugger
        return debugger

    @_validate_arguments
    def install( self, alias: str = builtins_alias_default ) -> __.typx.Self:
        ''' Installs truck into builtins with provided alias.

            Replaces an existing truck. Preserves global module configurations.

            Library developers should call :py:func:`register_module` instead.
        '''
        import builtins
        with _installer_lock:
            truck_o = getattr( builtins, alias, None )
            if isinstance( truck_o, Truck ):
                self( 'note', module_name = __name__ )(
                    'Installed truck is being replaced.' )
                setattr( builtins, alias, self )
            else:
                __.install_builtin_safely(
                    alias, self, _exceptions.AttributeNondisplacement )
        return self

    @_validate_arguments
    def register_module(
        self,
        name: __.Absential[ str ] = __.absent,
        configuration: __.Absential[ _cfg.ModuleConfiguration ] = __.absent,
    ) -> __.typx.Self:
        ''' Registers configuration for module.

            If no module or package name is given, then the current module is
            inferred.

            If no configuration is provided, then a default is generated.
        '''
        if __.is_absent( name ):
            name = _discover_invoker_module_name( )
        if __.is_absent( configuration ):
            configuration = _cfg.ModuleConfiguration( )
        with _registrar_lock:
            self.modulecfgs[ name ] = configuration
        return self

InstallAliasArgument: __.typx.TypeAlias = __.typx.Annotated[
    str,
    __.typx.Doc(
        ''' Alias under which the truck is installed in builtins. ''' ),
]
ProduceTruckActiveFlavorsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ ActiveFlavorsLiberal | ActiveFlavorsRegistryLiberal ],
    __.typx.Doc(
        ''' Flavors to activate.

            Can be collection, which applies globally across all registered
            modules. Or, can be mapping of module names to sets.

            Module-specific entries merge with global entries.
        ''' ),
]
ProduceTruckEvnActiveFlavorsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ __.typx.Optional[ str ] ],
    __.typx.Doc(
        ''' Name of environment variable for active flavors or ``None``.

            If absent, then a default environment variable name is used.

            If ``None``, then active flavors are not parsed from the process
            environment.

            If active flavors are supplied directly to a function,
            which also accepts this argument, then active flavors are not
            parsed from the process environment.
        ''' ),
]
ProduceTruckEvnTraceLevelsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ __.typx.Optional[ str ] ],
    __.typx.Doc(
        ''' Name of environment variable for trace levels or ``None``.

            If absent, then a default environment variable name is used.

            If ``None``, then trace levels are not parsed from the process
            environment.

            If trace levels are supplied directly to a function,
            which also accepts this argument, then trace levels are not
            parsed from the process environment.
        ''' ),
]
ProduceTruckFlavorsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ _cfg.FlavorsRegistryLiberal ],
    __.typx.Doc( ''' Registry of flavor identifiers to configurations. ''' ),
]
ProduceTruckGeneralcfgArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ _cfg.VehicleConfiguration ],
    __.typx.Doc(
        ''' General configuration for the truck.

            Top of configuration inheritance hierarchy. If absent,
            defaults to a suitable configuration for application use.
        ''' ),
]
ProduceTruckModulecfgsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ ModulesConfigurationsRegistryLiberal ],
    __.typx.Doc(
        ''' Module configurations for the truck.

            If absent, defaults to global modules registry.
        ''' ),
]
ProduceTruckPrinterFactoryArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ _printers.PrinterFactoryUnion ],
    __.typx.Doc(
        ''' Factory which produces callables to output text somewhere.

            May also be writable text stream.
            Factories take two arguments, module name and flavor, and
            return a callable which takes one argument, the string
            produced by a formatter.

            If absent, uses a default.
        ''' ),
]
ProduceTruckTraceLevelsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ int | TraceLevelsRegistryLiberal ],
    __.typx.Doc(
        ''' Maximum trace depths.

            Can be an integer, which applies globally across all registered
            modules. Or, can be a mapping of module names to integers.

            Module-specific entries override global entries.
        ''' ),
]
RegisterModuleFormatterFactoryArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ _cfg.FormatterFactory ],
    __.typx.Doc(
        ''' Factory which produces formatter callable.

            Takes formatter control, module name, and flavor as arguments.
            Returns formatter to convert an argument to a string.
        ''' ),
]
RegisterModuleIncludeContextArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ bool ],
    __.typx.Doc( ''' Include stack frame with output? ''' ),
]
RegisterModuleNameArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ str ],
    __.typx.Doc(
        ''' Name of the module to register.

            If absent, infers the current module name.
        ''' ),
]
RegisterModulePrefixEmitterArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ _cfg.PrefixEmitterUnion ],
    __.typx.Doc(
        ''' String or factory which produces output prefix string.

            Factory takes formatter control, module name, and flavor as
            arguments. Returns prefix string.
        ''' ),
]


def active_flavors_from_environment(
    evname: __.Absential[ str ] = __.absent
) -> ActiveFlavorsRegistry:
    ''' Extracts active flavors from named environment variable. '''
    active_flavors: ActiveFlavorsRegistryLiberal = { }
    name = 'ICTRUCK_ACTIVE_FLAVORS' if __.is_absent( evname ) else evname
    value = __.os.getenv( name, '' )
    for part in value.split( '+' ):
        if not part: continue
        if ':' in part:
            mname, flavors = part.split( ':', 1 )
        else: mname, flavors = None, part
        match flavors:
            case '*': active_flavors[ mname ] = omniflavor
            case _: active_flavors[ mname ] = flavors.split( ',' )
    return __.immut.Dictionary( {
        mname:
            flavors if isinstance( flavors, Omniflavor )
            else frozenset( flavors )
        for mname, flavors in active_flavors.items( ) } )


def trace_levels_from_environment(
    evname: __.Absential[ str ] = __.absent
) -> TraceLevelsRegistry:
    ''' Extracts trace levels from named environment variable. '''
    trace_levels: TraceLevelsRegistryLiberal = { None: -1 }
    name = 'ICTRUCK_TRACE_LEVELS' if __.is_absent( evname ) else evname
    value = __.os.getenv( name, '' )
    for part in value.split( '+' ):
        if not part: continue
        if ':' in part: mname, level = part.split( ':', 1 )
        else: mname, level = None, part
        if not level.isdigit( ):
            __.warnings.warn(
                f"Non-integer trace level {level!r} "
                f"in environment variable {name!r}." )
            continue
        trace_levels[ mname ] = int( level )
    return __.immut.Dictionary( trace_levels )


@_validate_arguments
def install( # noqa: PLR0913
    alias: InstallAliasArgument = builtins_alias_default,
    active_flavors: ProduceTruckActiveFlavorsArgument = __.absent,
    generalcfg: ProduceTruckGeneralcfgArgument = __.absent,
    printer_factory: ProduceTruckPrinterFactoryArgument = __.absent,
    trace_levels: ProduceTruckTraceLevelsArgument = __.absent,
    evname_active_flavors: ProduceTruckEvnActiveFlavorsArgument = __.absent,
    evname_trace_levels: ProduceTruckEvnTraceLevelsArgument = __.absent,
) -> Truck:
    ''' Produces truck and installs it into builtins with alias.

        Replaces an existing truck, preserving global module configurations.

        Library developers should call :py:func:`register_module` instead.
    '''
    truck = produce_truck(
        active_flavors = active_flavors,
        generalcfg = generalcfg,
        printer_factory = printer_factory,
        trace_levels = trace_levels,
        evname_active_flavors = evname_active_flavors,
        evname_trace_levels = evname_trace_levels )
    return truck.install( alias = alias )


@_validate_arguments
def produce_truck( # noqa: PLR0913
    active_flavors: ProduceTruckActiveFlavorsArgument = __.absent,
    generalcfg: ProduceTruckGeneralcfgArgument = __.absent,
    modulecfgs: ProduceTruckModulecfgsArgument = __.absent,
    printer_factory: ProduceTruckPrinterFactoryArgument = __.absent,
    trace_levels: ProduceTruckTraceLevelsArgument = __.absent,
    evname_active_flavors: ProduceTruckEvnActiveFlavorsArgument = __.absent,
    evname_trace_levels: ProduceTruckEvnTraceLevelsArgument = __.absent,
) -> Truck:
    ''' Produces icecream truck with some shorthand argument values. '''
    # TODO: Deeper validation of active flavors and trace levels.
    # TODO: Deeper validation of printer factory.
    initargs: dict[ str, __.typx.Any ] = { }
    if not __.is_absent( generalcfg ):
        initargs[ 'generalcfg' ] = generalcfg
    if not __.is_absent( modulecfgs ):
        initargs[ 'modulecfgs' ] = ModulesConfigurationsRegistry(
            {   mname: configuration for mname, configuration
                in modulecfgs.items( ) } )
    if not __.is_absent( printer_factory ):
        initargs[ 'printer_factory' ] = printer_factory
    _add_truck_initarg_active_flavors(
        initargs, active_flavors, evname_active_flavors )
    _add_truck_initarg_trace_levels(
        initargs, trace_levels, evname_trace_levels )
    return Truck( **initargs )


@_validate_arguments
def register_module(
    name: RegisterModuleNameArgument = __.absent,
    flavors: ProduceTruckFlavorsArgument = __.absent,
    formatter_factory: RegisterModuleFormatterFactoryArgument = __.absent,
    include_context: RegisterModuleIncludeContextArgument = __.absent,
    prefix_emitter: RegisterModulePrefixEmitterArgument = __.absent,
) -> _cfg.ModuleConfiguration:
    ''' Registers module configuration on the builtin truck.

        If no truck exists in builtins, installs one which produces null
        printers.

        Intended for library developers to configure debugging flavors
        without overriding anything set by the application or other libraries.
        Application developers should call :py:func:`install` instead.
    '''
    import builtins
    truck = getattr( builtins, builtins_alias_default, None )
    if not isinstance( truck, Truck ):
        truck = Truck( )
        __.install_builtin_safely(
            builtins_alias_default,
            truck,
            _exceptions.AttributeNondisplacement )
    nomargs: dict[ str, __.typx.Any ] = { }
    if not __.is_absent( flavors ):
        nomargs[ 'flavors' ] = __.immut.Dictionary( flavors )
    if not __.is_absent( formatter_factory ):
        nomargs[ 'formatter_factory' ] = formatter_factory
    if not __.is_absent( include_context ):
        nomargs[ 'include_context' ] = include_context
    if not __.is_absent( prefix_emitter ):
        nomargs[ 'prefix_emitter' ] = prefix_emitter
    configuration = _cfg.ModuleConfiguration( **nomargs )
    return truck.register_module( name = name, configuration = configuration )


def _add_truck_initarg_active_flavors(
    initargs: dict[ str, __.typx.Any ],
    active_flavors: ProduceTruckActiveFlavorsArgument = __.absent,
    evname_active_flavors: ProduceTruckEvnActiveFlavorsArgument = __.absent,
) -> None:
    name = 'active_flavors'
    if not __.is_absent( active_flavors ):
        if isinstance( active_flavors, Omniflavor ):
            initargs[ name ] = __.immut.Dictionary(
                { None: active_flavors } )
        elif isinstance( active_flavors, ( __.cabc.Sequence,  __.cabc.Set ) ):
            initargs[ name ] = __.immut.Dictionary(
                { None: frozenset( active_flavors ) } )
        else:
            initargs[ name ] = __.immut.Dictionary( {
                mname:
                    flavors if isinstance( flavors, Omniflavor )
                    else frozenset( flavors )
                for mname, flavors in active_flavors.items( ) } )
    elif evname_active_flavors is not None:
        initargs[ name ] = (
            active_flavors_from_environment( evname = evname_active_flavors ) )


def _add_truck_initarg_trace_levels(
    initargs: dict[ str, __.typx.Any ],
    trace_levels: ProduceTruckTraceLevelsArgument = __.absent,
    evname_trace_levels: ProduceTruckEvnTraceLevelsArgument = __.absent,
) -> None:
    name = 'trace_levels'
    if not __.is_absent( trace_levels ):
        if isinstance( trace_levels, int ):
            initargs[ name ] = __.immut.Dictionary( { None: trace_levels } )
        else:
            trace_levels_: TraceLevelsRegistryLiberal = { None: -1 }
            trace_levels_.update( trace_levels )
            initargs[ name ] = __.immut.Dictionary( trace_levels_ )
    elif evname_trace_levels is not None:
        initargs[ name ] = (
            trace_levels_from_environment( evname = evname_trace_levels ) )


def _calculate_effective_flavors(
    flavors: ActiveFlavorsRegistry, mname: str
) -> ActiveFlavors:
    result_ = flavors.get( None ) or frozenset( )
    if isinstance( result_, Omniflavor ): return result_
    result = result_
    for mname_ in _iterate_module_name_ancestry( mname ):
        if mname_ in flavors:
            result_ = flavors.get( mname_ ) or frozenset( )
            if isinstance( result_, Omniflavor ): return result_
            result |= result_
    return result


def _calculate_effective_trace_level(
    levels: TraceLevelsRegistry, mname: str
) -> int:
    result = levels.get( None, -1 )
    for mname_ in _iterate_module_name_ancestry( mname ):
        if mname_ in levels:
            result = levels[ mname_ ]
    return result


def _calculate_ic_initargs(
    truck: Truck,
    configuration: __.immut.Dictionary[ str, __.typx.Any ],
    control: _cfg.FormatterControl,
    mname: str,
    flavor: _cfg.Flavor,
) -> dict[ str, __.typx.Any ]:
    nomargs: dict[ str, __.typx.Any ] = { }
    nomargs[ 'argToStringFunction' ] = (
        configuration[ 'formatter_factory' ]( control, mname, flavor ) )
    nomargs[ 'includeContext' ] = configuration[ 'include_context' ]
    if isinstance( truck.printer_factory, __.io.TextIOBase ):
        printer = __.funct.partial( print, file = truck.printer_factory )
    else: printer = truck.printer_factory( mname, flavor )
    nomargs[ 'outputFunction' ] = printer
    prefix_emitter = configuration[ 'prefix_emitter' ]
    nomargs[ 'prefix' ] = (
        prefix_emitter if isinstance( prefix_emitter, str )
        else prefix_emitter( mname, flavor ) )
    return nomargs


def _dict_from_dataclass( objct: object ) -> dict[ str, __.typx.Any ]:
    # objct = __.typx.cast( _typeshed.DataclassInstance, objct )
    return {
        field.name: getattr( objct, field.name )
        for field in __.dcls.fields( objct ) # pyright: ignore[reportArgumentType]
        if not field.name.startswith( '_' ) }


def _discover_invoker_module_name( ) -> str:
    frame = __.inspect.currentframe( )
    while frame: # pragma: no branch
        module = __.inspect.getmodule( frame )
        if module is None:
            if '<stdin>' == frame.f_code.co_filename: # pragma: no cover
                name = '__main__'
                break
            raise _exceptions.ModuleInferenceFailure
        name = module.__name__
        if not name.startswith( f"{__.package_name}." ): break
        frame = frame.f_back
    return name


def _iterate_module_name_ancestry( name: str ) -> __.cabc.Iterator[ str ]:
    parts = name.split( '.' )
    for i in range( len( parts ) ):
        yield '.'.join( parts[ : i + 1 ] )


def _merge_ic_configuration(
    base: dict[ str, __.typx.Any ], update_objct: object,
) -> dict[ str, __.typx.Any ]:
    update: dict[ str, __.typx.Any ] = _dict_from_dataclass( update_objct )
    result: dict[ str, __.typx.Any ] = { }
    result[ 'flavors' ] = (
            dict( base.get( 'flavors', dict( ) ) )
        |   dict( update.get( 'flavors', dict( ) ) ) )
    for ename in ( 'formatter_factory', 'include_context', 'prefix_emitter' ):
        uvalue = update.get( ename )
        if uvalue is not None: result[ ename ] = uvalue
        elif ename in base: result[ ename ] = base[ ename ]
    return result


def _produce_ic_configuration(
    vehicle: Truck, mname: str, flavor: _cfg.Flavor
) -> __.immut.Dictionary[ str, __.typx.Any ]:
    fconfigs: list[ _cfg.FlavorConfiguration ] = [ ]
    vconfig = vehicle.generalcfg
    configd: dict[ str, __.typx.Any ] = {
        field.name: getattr( vconfig, field.name )
        for field in __.dcls.fields( vconfig )
        if not field.name.startswith( '_' ) }
    if flavor in vconfig.flavors:
        fconfigs.append( vconfig.flavors[ flavor ] )
    for mname_ in _iterate_module_name_ancestry( mname ):
        if mname_ not in vehicle.modulecfgs: continue
        mconfig = vehicle.modulecfgs[ mname_ ]
        configd = _merge_ic_configuration( configd, mconfig )
        if flavor in mconfig.flavors:
            fconfigs.append( mconfig.flavors[ flavor ] )
    if not fconfigs: raise _exceptions.FlavorInavailability( flavor )
    # Apply collected flavor configs after general and module configs.
    # (Applied in top-down order for correct overrides.)
    for fconfig in fconfigs:
        configd = _merge_ic_configuration( configd, fconfig )
    return __.immut.Dictionary( configd )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Common names and type aliases. '''


from . import imports as __


H = __.typx.TypeVar( 'H', bound = __.cabc.Hashable ) # Hash Key
V = __.typx.TypeVar( 'V' ) # Value


ComparisonResult: __.typx.TypeAlias = bool | __.types.NotImplementedType
NominativeArguments: __.typx.TypeAlias = __.cabc.Mapping[ str, __.typx.Any ]
PositionalArguments: __.typx.TypeAlias = __.cabc.Sequence[ __.typx.Any ]


DictionaryNominativeArgument: __.typx.TypeAlias = __.typx.Annotated[
    V,
    __.ddoc.Doc(
        'Zero or more keyword arguments from which to initialize '
        'dictionary data.' ),
]
DictionaryPositionalArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.cabc.Mapping[ H, V ] | __.cabc.Iterable[ tuple[ H, V ] ],
    __.ddoc.Doc(
        'Zero or more iterables from which to initialize dictionary data. '
        'Each iterable must be dictionary or sequence of key-value pairs. '
        'Duplicate keys will result in an error.' ),
]
ExceptionInfo: __.typx.TypeAlias = tuple[
    type[ BaseException ] | None,
    BaseException | None,
    __.types.TracebackType | None ]


package_name = __name__.split( '.', maxsplit = 1 )[ 0 ]



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Flexible factory for Icecream debuggers.

    A debugging library that enhances the `icecream
    <https://github.com/gruns/icecream>`_ package with flexible, flavorful
    traces and module-specific configurations. Designed for both application
    and library developers, it provides granular control over debug output
    while ensuring isolation between different configurations.

    The package organizes its functionality across several modules, providing
    exceptions, configuration hierarchies, and specialized output recipes.
'''


from . import __
# --- BEGIN: Injected by Copier ---
from . import exceptions
# --- END: Injected by Copier ---


from .configuration import *
from .exceptions import *
from .printers import *
from .vehicles import *


__version__: str
__version__ = '1.5'


__.immut.finalize_module( __name__, recursive = True )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Recipes for integration with standard logging. '''



import logging as _logging

from . import __


_validate_arguments = (
    __.validate_arguments(
        globalvars = globals( ),
        errorclass = __.exceptions.ArgumentClassInvalidity ) )


InstallAdditionalAliasesArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ __.cabc.Mapping[ str, str ] ],
    __.typx.Doc(
        ''' Mapping of builtins attribute names to flavors. ''' ),
]


@_validate_arguments
def install(
    alias: __.InstallAliasArgument = __.builtins_alias_default,
    additional_aliases: InstallAdditionalAliasesArgument = __.absent,
) -> __.Truck:
    ''' Produces truck and installs it into builtins with alias.

        Replaces an existing truck, preserving global module configurations.

        Library developers should call :py:func:`__.register_module` instead.
    '''
    truck = produce_truck( ).install( alias = alias )
    if __.is_absent( additional_aliases ): additional_aliases = { }
    for falias, flavor in additional_aliases.items( ):
        __.install_builtin_safely(
            falias, truck( flavor ), __.exceptions.AttributeNondisplacement )
    return truck


@_validate_arguments
def produce_printer( mname: str, flavor: __.Flavor ) -> __.Printer:
    ''' Produces printer which maps flavors to logging levels. '''
    logger = _logging.getLogger( mname )
    level = (
        getattr( _logging, flavor.upper( ) ) if isinstance( flavor, str )
        else _logging.DEBUG )
    return lambda x: logger.log( level, x )


@_validate_arguments
def produce_truck( ) -> __.Truck:
    ''' Produces icecream truck which integrates with standard logging. '''
    active_flavors = { None: frozenset( {
        'debug', 'info', 'warning', 'error', 'critical' } ) }
    flavors: __.immut.Dictionary[ __.Flavor, __.FlavorConfiguration ] = (
        __.immut.Dictionary(
            {   name: __.FlavorConfiguration( )
                for name in active_flavors[ None ] } ) )
    generalcfg = __.VehicleConfiguration( flavors = flavors )
    nomargs = dict(
        active_flavors = active_flavors,
        generalcfg = generalcfg,
        printer_factory = produce_printer )
    return __.Truck( **nomargs ) # pyright: ignore
